"""
Discrete probability distributions
"""

import random

class DDist:
    """
    Discrete probability distribution.  Can be sparse, in the sense that
    elements that are not explicitly contained in the dictionary are assumed to
    have zero probability.
    """
    def __init__(self, elts):
        """
        :param elts: A dictionary mapping support elements to their associated
                     probabilities.  The values of this dictionary (the
                     probabilities) must be nonnegative, and they must sum to
                     1.
        """
        if not (abs(sum(elts.values())-1) < 1e-6 and min(elts.values()) >= 0.0):
            raise Exception("Probabilities must be nonnegative, and must sum to 1")
        self.d = elts

    def prob(self, elt):
        """
        :param elt: an element of the domain of this distribution
        :returns: the probability associated with ``elt``
        """
        return self.d.get(elt,0)

    def support(self):
        """
        :returns: A list (in arbitrary order) of the elements of this
                  distribution with non-zero probabability.
        """
        return [k for k in list(self.d.keys()) if self.prob(k) > 0]

    def draw(self):
        """
        :returns: A randomly drawn element from the distribution
        """
        r = random.random()
        sum = 0.0
        for val in self.support():
            sum += self.prob(val)
            if r < sum:
                return val

    def __repr__(self):# hartz, 13 june 2012
        return "DDist(%s)" % repr(self.d)

    __str__ = __repr__

    def expectation(self,f): # hartz, 13 june 2012
        """
        :returns: The expected value of function ``f``, with respect to this
                  distribution
        """
        return sum(self.prob(i)*f(i) for i in self.support())

    def max_prob_elt(self):
        """
        :returns: The element in this domain with maximum probability
        """
        return max(self.support(), key=self.prob)

    def project(self, mapFunc):
        """
        :param mapFunc: a function mapping elements in the support
                        to modified elements

        :returns: a ``DDist`` over some new set of elements, where the
                  probability associated with each element :math:`e` in the
                  original distribution is associated with :math:`e'` =
                  :math:`\\mathtt{mapFunc}(e)` in the new distribution.
        """
        result = {}
        for e in self.support():
            new = mapFunc(e)
            result[new] = result.get(new,0) + self.prob(e)
        return DDist(result)

    def condition(self, testFunc):
        """
        :param testFunc: a function mapping support elements to booleans:
                         ``True`` if the element should be included in the
                         resulting distribution, and ``False`` otherwise.

        :returns: new distribution, conditioned on the event described by
                  ``testfunc``
        """
        newElements = [e for e in self.support() if testFunc(e)]
        if len(newElements) == 0:
            raise ZeroDivisionError("division by zero (conditioning on impossible event)")
        z = sum(self.prob(e) for e in newElements)
        result = {}
        for e in newElements:
            result[e] = result.get(e,0) + self.prob(e)/z
        return DDist(result)

def bayes_rule(PA, prob_b_given_a, b):
    """
    :param PA: :py:class:`DDist` over :math:`A`
    :param prob_b_given_a: conditional distribution over :math:`B` given :math:`A`
                 (function from values of :math:`a` to :py:class:`DDist` over
                 :math:`B`)
    :param b: evidence value for :math:`B = b`
    :returns: Updated distribution given evidence, :math:`\\Pr(A~|~B=b)`
    """
    return make_joint_distribution(PA, prob_b_given_a).condition(lambda x: x[1]==b).\
        project(lambda x: x[0])

def make_joint_distribution(PA, prob_b_given_a):
    """
    Create a joint distribution on :math:`\\Pr(A,~B)` (in that order),
    represented as a :py:class:`DDist`

    :param PA: a :py:class:`DDist` on some random var :math:`A`
    :param prob_b_given_a: a conditional probability distribution specifying
                 :math:`\\Pr(B~|~A)` (that is, a function from elements of
                 :math:`A` to :py:class:`DDist` on :math:`B`)
    """
    d = {}
    for a in PA.support():
        for b in prob_b_given_a(a).support():
            d[(a, b)] = PA.prob(a) * prob_b_given_a(a).prob(b)
    return DDist(d)

def total_probability(PA, prob_b_given_a):
    """
    :param prob_b_given_a: conditional distribution over :math:`B` given :math:`A` (function
                 from values of :math:`a` to :py:class:`DDist` over :math:`B`)
    :param PA: distribution over :math:`A` (object of type :py:class:`DDist`)
    :returns: :math:`\\Pr(B)`, computed using the law of total probability.
    """
    #make_joint_distribution will make P(A,B), in that order.
    #want to marginalize out A, so get rid of index 0
    return make_joint_distribution(PA, prob_b_given_a).project(lambda x: x[1])

######################################################################
#   Special discrete of distributions

def delta_dist(v):
    """
    Distribution with all of its probability mass on value ``v``

    :param v: the element that should have probability 1
    """
    return DDist({v:1.0})

def uniform_dist(elts):
    """
    Uniform distribution over a given finite set of ``elts``

    :param elts: list of any kind of item
    """
    p = 1.0 / len(elts)
    return DDist({e:p for e in elts})

def mixture(d1,d2,p): # hartz, 13 june 2012
    """
    A mixture of two :py:class:`DDist` instances, ``d1`` and ``d2``, with
    mixture parameter ``p``.  Probability of an
    element :math:`x` under this distribution is :math:`p \\cdot \\mathtt{d1.prob}(x) + (1 - p) \\cdot \\mathtt{d2.prob}(x)`.
    """
    return DDist({k:d1.prob(k)*p + d2.prob(k)*(1-p) for k in set(d1.support()+d2.support())})

def triangle_dist(peak, halfWidth, loLimit = None, hiLimit = None):
    """
    Construct and return a DDist over integers. The distribution will have its
    peak at index ``peak`` and fall off linearly from there, reaching 0 at an
    index ``halfWidth`` on either side of ``peak``.  Any probability mass that
    would be below ``loLimit`` or above ``hiLimit`` is instead assigned to
    ``loLimit`` or ``hiLimit``
    """
    lo = loLimit if loLimit is not None else float('-inf')
    hi = hiLimit if hiLimit is not None else float('inf')
    d = {}
    d[max(lo, min(hi, peak))] = 1
    total = 1
    fhw = float(halfWidth)
    for offset in range(1, halfWidth):
        l = max(lo, min(hi, peak-offset))
        h = max(lo, min(hi, peak+offset))
        p = (halfWidth - offset) / fhw
        d[h] = d.get(h,0) + p
        d[l] = d.get(l,0) + p
        total += 2 * p
    for (elt, value) in list(d.items()):
        d[elt] = value / total
    return DDist(d)

def square_dist(lo, hi, loLimit = None, hiLimit = None):
    """
    Construct and return a DDist over integers.  The distribution will have a
    uniform distribution on integers from ``lo`` to ``hi-1`` (inclusive).  Any
    probability mass that would be below ``loLimit`` or above ``hiLimit`` is
    instead assigned to ``loLimit`` or ``hiLimit``.
    """
    d = {}
    p = 1.0 / (hi - lo)
    for i in range(lo, hi):
        actual = max((loLimit if loLimit is not None else float('-inf')), min((hiLimit if hiLimit is not None else float('inf')), i))
        d[actual] = d.get(actual,0) + p
    return DDist(d)
