"""
Framework for modeling and simulating Linear Time Invariant systems.
"""
from lib601.poly import Polynomial


def _fixnum(x):
    try:
        return float(x)
    except:
        return x

def _strip_all_zeros(num, denom):
    n, d = num[:], denom[:]
    while n and n[0] == 0:
        n.pop(0)
    while d and d[0] == 0:
        d.pop(0)
    while n and d and n[-1] == d[-1] == 0:
        n.pop(-1)
        d.pop(-1)
    return (n, d)

def _normalize(numerator, denominator):
    num, denom = _strip_all_zeros(numerator, denominator)
    norm_val = num[0]
    n = [_fixnum(i)/norm_val for i in num]
    d = [_fixnum(i)/norm_val for i in denom]
    return (num, denom)

def _approx_equal(x, y, epsilon=1e-4):
    return abs(x - y) < epsilon

def _approx_equal_list(l1, l2):
    return len(l1) == len(l2) and all(map(_approx_equal, *zip(l1, l2)))

def _sf_to_diffeq(numerator, denominator):
    numCoeffs = numerator._coeffs()
    denomCoeffs = denominator._coeffs()
    iCoeffs = [n/denominator.coeff(0) for n in numCoeffs]
    oCoeffs = [-d/denominator.coeff(0) for d in denomCoeffs[1:]]
    return (iCoeffs, oCoeffs)


class System:
    """
    Represents Linear Time Invariant systems, and supports step-by-step simulation.
    """
    def __init__(self, numerator, denominator=None, previous_inputs=None, previous_outputs=None):
        """
        :param numerator: a list of numbers representing the numerator of the system functional, 
                starting with lowest order coefficient
        :param denominator: a list of numbers representing the denominator of the system functional, 
                starting with lowest order coefficient
        :param previous_inputs: a list of previous input numbers to store in the initial state
        :param previous_outputs: a list of previous output numbers to store in the initial state

        :ivar numerator: list of numbers representing the numerator of the system functional
        :ivar denominator: list of numbers representing the denominator of the system functional
        :ivar initial_state: tuple of two lists representing the previous inputs and outputs 
            to the system
        """
        if isinstance(numerator, Polynomial):
            self.numerator = numerator
        else:
            self.numerator = Polynomial(numerator)

        if isinstance(denominator, Polynomial):
            self.denominator = denominator
        elif denominator is None:
            self.denominator = Polynomial([1])
        else:
            self.denominator = Polynomial(denominator)

        self.iCoeffs, self.oCoeffs = _sf_to_diffeq(self.numerator, self.denominator)

        previous_inputs = previous_inputs if previous_inputs is not None else [0]*len(self.iCoeffs)
        previous_outputs = previous_outputs if previous_outputs is not None else [0]*len(self.oCoeffs)

        self.initial_state = (previous_inputs, previous_outputs)

    def calculate_step(self, state, inp):
        """
        :param state: object representing a current hypothetical state of the system
        :param inp: input value to the system
        :returns: a tuple containing the system's new state and output after a hypothetical step
                using the given state and input. 
        Note that this is a pure function which does not store or modify any class, instance, or global variables.
        """
        inputs, outputs = state

        # update the list of inputs to include x[n] at the start
        inputs = [inp] + inputs

        output = sum(x*d for x,d in zip(inputs, self.iCoeffs))
        output += sum(y*c for y,c in zip(outputs, self.oCoeffs))

        # update the inputs and outputs for our state
        inputs = inputs[:-1]
        outputs = [output] + outputs[:-1]
        state = (inputs, outputs)
        return (state, output)

    def poles(self):
        """
        :returns: a list containing the poles of this system.
        """
        # Find the poles of a system functional by substituting
        # 1/z for R in both numerator and denominator, and then
        # finding the roots of the denominator polynomial of z.

        # We can find the coefficients of the z polynomial by reversing the
        # order of the coefficients of the R polynomial.
        # However, we must make the numerator and denominator coefficient lists
        # have the same length (by padding with zeros) to avoid introducing
        # false roots (at zero).

        # create equal-length lists of coefficients corresponding to polynomials in z
        num = []
        den = []
        for i in range(max(self.numerator.order,self.denominator.order), -1, -1):
            num.append(self.numerator.coeff(i))
            den.append(self.denominator.coeff(i))
        # cancel poles at zero with zeros at zero
        while len(num) > 0 and num[0]==den[0]==0:
            num.pop(0)
            den.pop(0)
        return Polynomial(den).roots()

    def dominant_pole(self):
        """
        :returns: the dominant pole of this system.
        """
        p = self.poles()
        if len(p) == 0:
            return None
        return max(p, key=abs)

    def __eq__(self, other):
        if type(self) != type(other):
            return False

        my_num, my_denom = _normalize(self.numerator, self.denominator)
        other_num, other_denom = _normalize(other.numerator, other.denominator)

        return _approx_equal_list(my_num, other_num) and _approx_equal_list(my_denom, other_denom)

    def __neq__(self, other):
        return not self.__eq__(other)


class SystemSimulator:
    """
    Simulates Linear Time Invariant systems, and supports resetting and stepping for individual or lists of inputs.
    """
    def __init__(self, system):
        """
        :param system: the instance of :py:class:`System` to simulate 

        :ivar system: the instance of :py:class:`System` to simulate 
        """
        self.system = system
        self.reset()

    def step(self, inp):
        """
        :param inp: input value to this simulator
        :returns: the output value of this simulator after executing with the given input
        Note that this function updates this simulator's internal state.
        """
        new_state, out = self.system.calculate_step(self.state, inp)
        self.state = new_state
        return out

    def reset(self):
        """
        Resets this simulator's state to ``self.system``'s initial state.
        """
        self.state = self.system.initial_state

    def get_response(self, inputs, reset=True):
        """
        :param inputs: a list of consecutive input values to this simulator
        :param reset: whether to reset this simulator prior to executing with these inputs
        :returns: a list of output values of the simulator after executing with the given inputs.
        Note that this function updates the system's internal state.
        """
        if reset:
            self.reset()
        return [self.step(inp) for inp in inputs]


class R(System):
    """
    Subclass of :py:class:`System` representing Delay systems. 
    Delay systems output at time :math:`n` the input passed into the system at time :math:`(n-1)`.
    """
    def __init__(self, output0=0):
        """
        :param output0: the value this system should output at time zero 

        :ivar initial_state: the initial state of this system 
        """
        self.initial_state = output0
        self.numerator = Polynomial([0, 1])
        self.denominator = Polynomial([1])

    def calculate_step(self, state, inp):
        """
        :param state: object representing a current hypothetical state of the system
        :param inp: input value to the system
        :returns: a tuple containing the system's new state and output after a hypothetical step
                using the given state and input. 
        Note that this is a pure function which does not store or modify any class, instance, or global variables.
        """
        out = state
        state = inp
        return (state, out)


class Gain(System):
    """
    Subclass of :py:class:`System` representing Gain systems. 
    Gain systems output at time :math:`n` the input passed into the system at time :math:`n`, multiplied by a gain ``k``.
    """
    def __init__(self, k):
        """
        :param k: the value this gain system should multiply inputs by 

        :ivar initial_state: the initial state of this system 
        :ivar k: the value this gain system should multiply inputs by
        """
        self.initial_state = None
        self.k = k
        self.numerator = Polynomial([k])
        self.denominator = Polynomial([1])

    def calculate_step(self, state, inp):
        """
        :param state: object representing a current hypothetical state of the system
        :param inp: input value to the system
        :returns: a tuple containing the system's new state and output after a hypothetical step
                using the given state and input. 
        Note that this is a pure function which does not store or modify any class, instance, or global variables.
        """
        out = self.k * inp
        return (state, out)


class Cascade(System):
    """
    Subclass of :py:class:`System` representing the Cascade composition of two :py:class:`System` s.
    Cascade systems pass input into their first subsystem, pass the output of their first subsystem as the input
    into the second subsystem, and finally output the second subsystem's output.
    """
    def __init__(self, s1, s2):
        """
        :param s1: the first :py:class:`System` in this cascade 
        :param s2: the second :py:class:`System` in this cascade

        :ivar initial_state: a tuple representing the initial states of ``self.s1`` and ``self.s2``
        :ivar s1: the first :py:class:`System` in this cascade 
        :ivar s2: the second :py:class:`System` in this cascade 
        """
        self.initial_state = (s1.initial_state, s2.initial_state)
        self.s1 = s1
        self.s2 = s2
        self.numerator = s1.numerator * s2.numerator
        self.denominator = s1.denominator * s2.denominator

    def calculate_step(self, state, inp):
        """
        :param state: a tuple representing current hypothetical states of ``self.s1`` and ``self.s2``
        :param inp: input value to the system
        :returns: a tuple containing the cascade system's new state (a tuple of ``self.s1``'s new state and 
                ``self.s2``'s new state) and output after a hypothetical step through each subsystem. 
        Note that this is a pure function which does not store or modify any class, instance, or global variables.
        """
        s1_state, s2_state = state
        s1_state, intermediate = self.s1.calculate_step(s1_state, inp)
        s2_state, out = self.s2.calculate_step(s2_state, intermediate)
        state = (s1_state, s2_state)
        return (state, out)


class FeedforwardAdd(System):
    """
    Subclass of :py:class:`System` representing the feedforward addition of two :py:class:`System` s.
    FeedforwardAdd systems pass the same input into each of their two subsystems, and output the sum of 
    their subsystems' outputs.
    """
    def __init__(self, s1, s2):
        """
        :param s1: the first :py:class:`System` in this feedforward addition 
        :param s2: the second :py:class:`System` in this feedforward addition

        :ivar initial_state: a tuple representing the initial states of ``self.s1`` and ``self.s2``
        :ivar s1: the first :py:class:`System` in this feedforward addition 
        :ivar s2: the second :py:class:`System` in this feedforward addition 
        """
        self.initial_state = (s1.initial_state, s2.initial_state)
        self.s1 = s1
        self.s2 = s2

        n1, d1 = self.s1.numerator, self.s1.denominator
        n2, d2 = self.s2.numerator, self.s2.denominator
        self.numerator = n1*d2 + n2*d1
        self.denominator = d1*d2

    def calculate_step(self, state, inp):
        """
        :param state: a tuple representing current hypothetical states of ``self.s1`` and ``self.s2``
        :param inp: input value to the system
        :returns: a tuple containing the FeedforwardAdd system's new state (a tuple of ``self.s1``'s new state and 
                ``self.s2``'s new state) and output after a hypothetical step through each subsystem. 
        Note that this is a pure function which does not store or modify any class, instance, or global variables.
        """
        s1_state, s2_state = state
        s1_state, out1 = self.s1.calculate_step(s1_state, inp)
        s2_state, out2 = self.s2.calculate_step(s2_state, inp)
        state = (s1_state, s2_state)
        out = out1 + out2
        return (state, out)


class FeedbackAdd(System):
    """
    Subclass of :py:class:`System` representing the feedback addition of two :py:class:`System` s.
    In FeedbackAdd systems, the second subsystem's output is added to the input to the FeedbackAdd system. This sum is passed as 
    input into the first subsystem, whose consequent output is both the output of the overall system and the input
    back into the second subsystem. For this to work, we assume that at least one of the subsystems has output that 
    doesn't depend directly on its input.
    """
    def __init__(self, s1, s2):
        """
        :param s1: the first :py:class:`System` in this feedback addition 
        :param s2: the second :py:class:`System` in this feedback addition

        :ivar initial_state: a tuple representing the initial states of ``self.s1`` and ``self.s2``
        :ivar s1: the first :py:class:`System` in this feedback addition 
        :ivar s2: the second :py:class:`System` in this feedback addition 
        """
        self.initial_state = (s1.initial_state, s2.initial_state)
        self.s1 = s1
        self.s2 = s2

        n1, d1 = self.s1.numerator, self.s1.denominator
        n2, d2 = self.s2.numerator, self.s2.denominator
        self.numerator = n1*d2
        self.denominator = d1*d2 - n1*n2

    def calculate_step(self, state, inp):
        """
        :param state: a tuple representing current hypothetical states of ``self.s1`` and ``self.s2``
        :param inp: input value to the system
        :returns: a tuple containing the cascade system's new state (a tuple of ``self.s1``'s new state and 
                ``self.s2``'s new state) and output after a hypothetical step through each subsystem. 
        Note that this is a pure function which does not store or modify any class, instance, or global variables.
        """
        s1_state, s2_state = state

        # Here we encounter a problem.  It seems like s1's output
        # depends on s2's output, and s2's output depends on s1's
        # output.  Which one do we run first?  It looks like getting
        # either output is impossible!
        #
        # This is why we made the assumption that at least one of s1
        # and s2 has output that doesn't depend directly on its input.
        # Let's assume that s1 satisfies that property.  That means
        # that (for a particular state) we can give s1 any input, and
        # it will always give us the same output.  However, the
        # resulting state may not be correct.
        #
        # If we could "hypothetically step" s1 to get its output, do
        # the same to get s2's output, and then "actually step" s1 and
        # s2 to get their correct states, everything would work out.
        # Turns out we can do just that!
        #
        # However, we don't know if it's s1 or s2 that contains a
        # delay in the right place, so we have to make sure we try
        # them both.  Either way, once we've made it through both
        # machines, we know the correct output from s2, which solves
        # our problem.

        # The hypothetical step.  This looks a lot like cascade, with
        # the exception that all of our results are marked as
        # hypothetical.
        s1_state_hyp, out1_hyp = self.s1.calculate_step(s1_state, inp)
        s2_state_hyp, out2_hyp = self.s2.calculate_step(s2_state, out1_hyp)

        # We said before that the real out2 and hypothetical out2 are
        # exactly the same.
        out2_real = out2_hyp

        # Now we step it for real
        s1_state_real, out1_real = self.s1.calculate_step(s1_state, inp + out2_real)
        s2_state_real, out2_real = self.s2.calculate_step(s2_state, out1_real)

        # The states that come out of the real step are the ones we
        # should keep.  The out1 from the real step is the one we send
        # out into the rest of the world.
        state = (s1_state_real, s2_state_real)
        out = out1_real

        return (state, out)

def contains_io_delay(sm):
    state = sm.initial_state
    out1 = self.calculate_step(state, 1)
    out2 = self.calculate_step(state, 2)
    return out1 == out2
