"""
Tools for working with Markov Chains and Hidden Markov Models
"""

from . import dist
import operator
from functools import reduce

class MarkovChain:
    """
    Class for describing a Markov chain and computing some useful properties of
    it.
    """
    def __init__(self, initial_distribution,
                 transition_model):
        """
        :param initial_distribution: :math:`\\Pr(S_0)`, the initial distribution
                                  over states, represented as an instance of
                                  :py:class:`lib601.dist.DDist`
        :param transition_model: :math:`\\Pr(S_{t+1}~|~S_t)` represented
                                       as a procedure that takes an old state
                                       and returns a
                                       :py:class:`lib601.dist.DDist` over new
                                       states.

        :ivar initial_distribution: The value passed in as ``initial_distribution``
        :ivar transition_model: The value passed in as
                                      ``transition_model``
        """
        self.initial_distribution = initial_distribution
        self.transition_model = transition_model  

    def sample_sequence(self, T):
        """
        Returns a sequence of length ``T``, drawn from the distribution over
        sequences
        """
        result = [self.initial_distribution.draw()]
        for i in range(T-1):
            result.append(self.transition_model(result[-1]).draw())
        return result
    
    def state_sequence_prob(self, seq):
        """
        Returns the probability of a sequence of states in this Markov Chain
        
        :param seq: list of states :math:`[s_0, s_1, \\ldots, s_T]`
        """
        return reduce(operator.mul,
                      [self.transition_model(seq[t]).prob(seq[t+1]) \
                       for t in range(len(seq)-1)],
                      self.initial_distribution.prob(seq[0]))

    def occupation_dist(self, T):
        """
        Returns a :py:class:`lib601.dist.DDist` over states at time ``T``

        :param T: time step
        """
        if T == 0:
            return self.initial_distribution
        else:
            return dist.total_probability(self.occupation_dist(T-1),
                                         self.transition_model)
    
    def make_simulator(self):
        """
        :returns: An instance of :py:class:`MarkovChainSimulator` for
                  simulating the behavior of this Markov Chain
        """
        return MarkovChainSimulator(self)

class MarkovChainSimulator:
    """
    Class for simulating the evolution of a system described by a Markov chain    
    """

    def __init__(self, mc):
        """
        :param mc: The instance of :py:class:`MarkovChain` to be simulated.

        :ivar mc: The instance of :py:class:`MarkovChain` that was passed in as ``mc``.
        :ivar state: Holds the "current" state of the system (will have value
                     ``None`` until :py:meth:`initialize` is called, and will
                     be updated by :py:meth:`transition`).
        """
        self.mc = mc
        self.state = None

    def initialize(self):
        """
        Randomly selects and element from :py:attr:`self.mc.initial_distribution` and stores the
        result in the instance variable :py:attr:`state`.  Returns ``None``.
        """
        self.state = self.mc.initial_distribution.draw()
        return self.state

    def transition(self):
        """
        Updates the instance variable :py:attr:`state` by randomly selecting an element
        from :math:`\\Pr(S_{t+1}~|~S_t=\\mathtt{self.state})`.  Also returns the new state.
        """
        self.state = self.mc.transition_model(self.state).draw()
        return self.state


class HMM:
    """
    Class for specifying a hidden Markov model
    """
    def __init__(self, initial_distribution,
                       transition_model,
                       observation_model):
        """
        :param initial_distribution: :math:`\\Pr(S_0)`, the initial distribution
                                  over states, represented as a
                                  :py:class:`lib601.dist.DDist`
        :param transition_model: :math:`\\Pr(S_{t+1}~|~S_t)` represented as a
                                       procedure that takes an old state and
                                       returns a :py:class:`lib601.dist.DDist`
                                       over new states.
        :param observation_model: :math:`\\Pr(O_t~|~S_t)` represented as
                                        a procedure that takes a state and
                                        returns a :py:class:`lib601.dist.DDist`
                                        over observations that can be made if
                                        the system is in that state.
        
        :ivar initial_distribution: The value passed in as ``initial_distribution``
        :ivar transition_model: The value passed in as
                                      ``transition_model``
        :ivar observation_model: The value passed in as
                                      ``observation_model``
        """
        self.initial_distribution = initial_distribution
        self.transition_model = transition_model  
        self.observation_model = observation_model

    def make_simulator(self):
        """
        :returns: An instance of :py:class:`HMMSimulator` for
                  simulating the behavior of this HMM
        """
        return HMMSimulator(self)

    def make_state_estimator(self):
        """
        :returns: An instance of :py:class:`StateEstimator` for
                  performing Bayesian State Estimation using the models
                  described by this HMM
        """
        return StateEstimator(self)


class StateEstimator:
    """
    Class for estimating the state of some underlying system whose behavior is
    described by a Hidden Markov Model.
    """

    def __init__(self, hmm):
        """
        :param hmm: An instance of :py:class:`HMM` describing the underlying
                    system

        :ivar hmm: The value passed in as ``hmm``
        :ivar belief: The "current" belief, as a :py:class:`lib601.dist.DDist`
                      over states.  This variable is initialized to
                      :py:attr:`hmm.initial_distribution` and is updated by
                      calling :py:meth:`reset`, :py:meth:`transition`,
                      :py:meth:`observe`, or :py:meth:`update`.
        """
        self.hmm = hmm
        self.reset()

    def reset(self):
        """
        Reset the estimator so that the instance variable :py:attr:`belief` is
        equal to the initial belief.
        """
        self.belief = self.hmm.initial_distribution

    def transition(self):
        """
        Update the belief following a transition. Changes :py:attr:`belief` but
        does not return anything.

        :returns: ``None``
        """
        self.belief = dist.total_probability(self.belief,
                                             self.hmm.transition_model)

    def observe(self, obs):
        """
        Update the belief based on making an observation.  Changes
        :py:attr:`belief` but does not return anything.

        :param obs: The observation that was made

        :returns: ``None``
        """
        self.belief = dist.bayes_rule(self.belief,
                                      self.hmm.observation_model,
                                      obs)

    def update(self, obs):
        """
        Update the belief based on making an observation and then
        transitioning.  Changes :py:attr:`belief` but does not return anything.
        Same effect as :py:meth:`transition` followed by :py:meth:`observe`,
        but more efficient.

        :param obs: The observation that was made
        :returns: ``None``
        """
        afterObs = {}
        for s in self.belief.support():
            afterObs[s] = self.belief.prob(s) * \
                          self.hmm.observation_model(s).prob(obs)

        z = sum(afterObs.values()) #normalization factor

        if z == 0:
            raise Exception('Observation %s has 0 probability in all possible states.' % obs)

        new = {}
        tDist = self.hmm.transition_model
        for sPrime in self.belief.support(): #for each old state sPrime
            tDistS = tDist(sPrime) #distribution over ending states
            oldP = afterObs[sPrime] / z #probability of being in old state
            for s in tDistS.support():
                #prob of transitioning to s from sPrime                
                new[s] = new.get(s,0) + tDistS.prob(s)*oldP 
        self.belief = dist.DDist(new)

class HMMSimulator:
    """Class for simulating the behavior of some system whose behavior is
    described by a Hidden Markov Model."""

    def __init__(self, hmm):
        self.hmm = hmm
        self.state = None
        """
        Holds the "current" state of the system (will have value C{None} until
        L{initialize} is called, and will be updated by L{transition}).
        """

    def initialize(self):
        """
        Randomly selects and element from C{initial_distribution} and stores the
        result in the instance variable L{state}.  Returns C{None}.
        """
        self.state = self.hmm.initial_distribution.draw()

    def make_observation(self):
        """
        Returns an observation randomly drawn from P(O_t | S_t), where S_t is
        the internal state of the system (stored in the instance variable
        L{state}).
        """
        return self.hmm.observation_model(self.state).draw()

    def transition(self):
        """
        Updates the instance variable L{state} by randomly selecting an element
        from P(S_(t+1) | S_(t)).  Returns C{None}.
        """
        self.state = self.hmm.transition_model(self.state).draw()

