"""
Procedures for finding values of a function to minimize its output.
"""

import operator

def linear(objective, xmin, xmax, numXsteps):
    """
    Find the local minimum of a function ``objective`` between ``xmin`` and
    ``xmax`` using a linear search.

    :param objective: A function that takes a single number ``x`` as an
                      argument and returns a numerical value.
    :param xmin: The minimum value of ``x`` to consider, inclusive.
    :param xmax: The maximum value of ``x`` to consider, inclusive.
    :param numXsteps: The number of "steps" to try.
    :returns: A tuple ``(xBest, fBest)``.  ``xBest`` is one of the numeric
               values achieved by starting at ``xmin`` and taking
               ``numXsteps`` equal-sized steps up to ``xmax``.  The
               particular value of ``x`` returned is the one for which
               ``objective(x)`` is minimized.
    """
    fBest = None
    xBest = None
    cur = float(xmin)
    step = (xmax-cur) / numXsteps
    while cur <= xmax:
        fx = objective(cur)
        if fBest is None or compare(fx, fBest):
            fBest = fx
            xBest = cur
        cur += step
    return (xBest, fBest)


def bisection(objective, xmin, xmax, threshold=1e-4, h=1e-4):
    """
    Find the local minimum of a function ``objective`` between ``xmin`` and ``xmax``
    using the bisection method.

    :param objective: a function from a single numeric parameter to a single
                      numeric output
    :param xmin: the minimum value of ``objective``'s parameter that should be
                 considered
    :param xmax: the maximum value of ``objective``'s parameter that should be
                 considered
    :param threshold: the desired precision of the answer (the "optimum"
                      parameter returned will be within ``threshold`` of the
                      theoretical optimum parameter)
    :param h: the "delta" to use when approximating the derivative of
              ``objective``
    :returns: a tuple ``(xBest,fBest)``, where ``xBest`` is the value of ``x``
              for which ``f(x)`` is smallest (accurate to within ``threshold``
              of the theoretical best ``x``), and ``fBest`` is
              ``objective(xBest)``
    """
    f = objective
    f_prime = lambda x: (f(x+h)-f(x-h))/(2.0*h) # approximate derivative
    right = (xmax,f_prime(xmax))
    left = (xmin,f_prime(xmin))
    while abs(right[0] - left[0]) > threshold: 
        new_x = ((right[0]+left[0])/2.)
        new_d = f_prime(new_x)
        if (new_d < 0 and left[1] < 0) or (new_d > 0 and left[1] > 0):
            left = (new_x,new_d)
        else:
            right = (new_x,new_d)
    return left[0], f(left[0])
