"""
Polynomials with addition, subtraction, multiplication, and roots.
"""

class Polynomial:
    """
    Represent polynomials, and supports addition, subtraction, multiplication,
    and root finding.
    """

    def __init__(self, c):
        """
        :param c: a list of numbers, starting with lowest order coefficient.

        :ivar order: Integer representing the order of the polynomial
        """
        self.coeffs = tuple(float(i) if not isinstance(i,complex) else i for i in c)
        while len(self.coeffs) > 0 and self.coeffs[-1]==0:
            self.coeffs = self.coeffs[:-1]
        if len(self.coeffs) == 0:
            self.coeffs = (0,)
            self.order = 0
        else:
            self.order = len(self.coeffs) - 1

    def coeff(self,i):
        """
        :returns: the coefficient associated with the :math:`x^{\\mathtt{i}}`
                  term of the polynomial
        """
        return self.coeffs[i] if 0<=i<=self.order else 0.0

    def _coeffs(self, length=None):
        length = max(0, length) if length is not None else self.order+1
        return [self.coeff(i) for i in range(length)]

    def val(self, v):
        """
        :param v: number
        :returns: the value of the polynomial when :math:`x` has the value
                  specified by the argument ``v``.
        """
        if len(self.coeffs) == 0:
            return 0.0
        res = self.coeffs[-1]
        for i in range(self.order - 1, -1, -1):
            res = res * v + self.coeff(i)
        return res

    def add(self, other):
        """
        :param other: the instance of :py:class:`Polynomial` which we want to
                      add to ``self``
        :returns: a new instance of :py:class:`Polynomial`, which is the sum of
                  ``self`` and ``other``

        Does not affect either input.
        """
        c1 = [self.coeff(i) for i in range(self.order+1)]
        c2 = [other.coeff(i) for i in range(other.order+1)]
        while len(c1) < len(c2):
            c1.append(0)
        while len(c2) < len(c1):
            c2.append(0)
        return Polynomial([i+j for (i,j) in zip(c1, c2)])

    def mul(self, other):
        """
        :param other: the instance of :py:class:`Polynomial` which we want to
                      multiply by ``self``
        :returns: a new instance of :py:class:`Polynomial`, which is the
                  product of ``self`` and ``other``

        Does not affect either input.
        """
        out = Polynomial([])
        c1 = [self.coeff(i) for i in range(self.order+1)]
        for i in range(other.order+1):
            m = other.coeff(i)
            coeffs = [m*j for j in ([0]*i + c1)]
            p = Polynomial(coeffs)
            out = out + p
        return out

    def sub(self,other):
        """
        :param other: the instance of :py:class:`Polynomial` which we want to
                      subtract from ``self``
        :returns: a new instance of :py:class:`Polynomial`, which is the
                  difference of ``self`` and ``other``

        Does not affect either input.
        """
        return self + Polynomial([-1]) * other

    def __add__(self, other):
        return self.add(other)

    def __sub__(self,other):
        return self.sub(other)

    def __mul__(self, other):
        return self.mul(other)

    def __repr__(self):
        return 'Polynomial([%s])' % ', '.join(repr(i) for i in self.coeffs)

    def __str__(self):
        if self.order == 0:
            return str(self.coeff(0))
        out = ''
        for i in range(self.order,-1,-1):
            c = self.coeff(i)
            if c == 0:
                continue
            if c.real >= 0 and len(out) != 0:
                out += ' + '
            elif len(out) != 0:
                out += ' - '
                c = -c
            if c != 1:
                out += '(%r)' % c
            elif c == 1 and i == 0:
                out += repr(c)
            if i == 1:
                out += 'x'
            elif i != 0:
                out += '(x^%d)' % i
        return out

    def roots(self, start=complex(.4,.9), epsilon=1e-12, max_iterations=1000):
        """
        Computes the roots of the Polynomial using the Durand-Kerner method.

        :param start: (optional) seed value, defaults to :math:`.4+.9j`.
                      Initial guesses for the :math:`k` roots are
                      :math:`{\mathtt start}^0, {\mathtt start}^1,
                      \ldots, {\mathtt start}^k`

        :param epsilon: (optional) threshold, defaults to ``1e-12``.
                        The algorithm terminates (returning the roots)
                        when one iteration changes none of the roots
                        by more than ``epsilon``.
        
        :param max_iterations: (optional) the maximum number of iterations
                               to run, defaults to ``1000``.  If this number
                               of iterations is reached and the algorithm has
                               not converged, it will terminate and return
                               ``None``.

        :returns: A list containing the roots of the polynomial
        """
        if self.order == 1:
            a, b = [self.coeff(i) for i in range(self.order,-1,-1)]
            return [-b/a]
        if self.order <= 2:
            a, b, c = [self.coeff(i) for i in range(self.order,-1,-1)]
            if a == 0:
               return [-c/b]
            else:
               det = (b**2 - 4*a*c)
               return [(-b+det**0.5) / (2*a), (-b-det**0.5) / (2*a)]
        norm_factor = self.coeff(self.order)
        normed = Polynomial([i/norm_factor for i in self.coeffs])
        roots = [start ** i for i in range(normed.order)]
        n = 0
        for i in range(max_iterations):
            deltas = [0 for i in roots]
            for ix, root in enumerate(roots):
                t = normed.val(root)
                for jx, root2 in enumerate(roots):
                    if jx == ix:
                        continue
                    t /= (root - root2)
                new = root - t
                deltas[ix] = abs(new - root)
                roots[ix] = new
            if all(i < epsilon for i in deltas):
                return roots
        return roots


def _clipper(n, epsilon):
    return complex(n.real if abs(n.real) > epsilon else 0,
                   n.imag if abs(n.imag) > epsilon else 0)

def _floater(n):
    return n.real if n.imag == 0 else n
