"""
Procedures and classes for doing breadth-first and depth-first search, and
uniform-cost search.
"""

import heapq

class SearchNode:
    """
    A class representing a single node in a search tree.
    """
    def __init__(self, state, parent, cost = 0.):
        """
        :param state: The search-space state represented by this node.
        :param parent: This node's parent, as an instance of
                       :py:class:`SearchNode`, or ``None`` if this node is the
                       root of the search tree.
        :param cost: The cost of the *entire path* from the root of the search
                     tree to this node.
        """
        self.state = state
        self.parent = parent
        self.cost = cost

    def path(self):
        """
        :returns: A list of states representing the partial path from the root
                  of the search tree to this node.
        """
        p = []
        node = self
        while node:
            p.append(node.state)
            node = node.parent
        return p[::-1]

    def __repr__(self):
        return '->'.join([str(state) for state in self.path()])

    def __lt__(self, other):
        if isinstance(other, SearchNode):
            cmp = other.cost
        else:
            cmp = other
        return self.cost < cmp

    def __eq__(self, other):
        if isinstance(other, SearchNode):
            cmp = other.cost
        else:
            cmp = other
        return self.cost == cmp

    def __hash__(self):
        return hash((self.state, self.cost, self.parent))


def search(successors, start_state, goal_test, dfs = False):
    """
    Run a basic (BFS or DFS) search from start_state, until ``goal_test`` is
    satisfied, using the dynamic programming principle.

    :param successors: A function defining the search domain; takes a state as
                       input and returns a list of the states that are directly
                       reachable from that state.

    :param start_state: The state from which to start the search.

    :param goal_test: A function whose input is a state, and which outputs
                     ``True`` if that state satisfies the goal conditions, and
                     ``False`` otherwise.

    :param dfs: An optional boolean argument.  If ``True``, will run
                depth-first search.  If ``False`` (default), will run
                breadth-first search.

    :returns: A list of states, beginning with ``start_state``, representing the
              path; or ``None`` if no path from start to goal exists.
    """
    if goal_test(start_state):
        return [start_state]
    else:
        agenda = [SearchNode(start_state, None)]
        visited = {start_state}
        while len(agenda) > 0:
            if dfs:
                parent = agenda.pop(-1)
            else:
                parent = agenda.pop(0)
            for child_state in successors(parent.state):
                child = SearchNode(child_state, parent)
                if goal_test(child_state):
                    return child.path()
                if child_state not in visited:
                    agenda.append(child)
                    visited.add(child_state)
        return None


def depth_first_search(successors, start_state, goal_test):
    return search(successors, start_state, goal_test, dfs = True)


def breadth_first_search(successors, start_state, goal_test):
    return search(successors, start_state, goal_test, dfs = False)


class PriorityQueue:
    def __init__(self, items=None):
        self.items = [] if items is None else items
        heapq.heapify(self.items)

    def push(self, item, priority):
        heapq.heappush(self.items, (priority, item))

    def pop(self):
        return heapq.heappop(self.items)

    def __len__(self):
        return self.items.__len__()

    def __str__(self):
        return "PriorityQueue(%r)" % self.items

def uniform_cost_search(successors, start_state, goal_test, heuristic=lambda s: 0):
    """
    Run a uniform-cost search from start_state, until ``goal_test`` is satisfied.

    :param successors: A function defining the search domain; takes a state as
                       input and returns a list of pairs of the states that are
                       directly reachable from this state with the associated
                       cost.  The list returned by this function should be a
                       list of ``(child, cost)`` tuples, where ``child`` is a
                       state that is directly reachable from the input state,
                       and ``cost`` is the cost of transitioning directly from
                       the state being considered to ``child``.

    :param start_state: The state from which to start the search.

    :param goal_test: A function whose input is a state, and which outputs
                     ``True`` if that state satisfies the goal conditions, and
                     ``False`` otherwise.

    :param heuristic: A function mapping states to numerical heuristic values.
                      Defaults to a function that always returns 0.

    :returns:  A list of states (beginning with ``start_state``), representing
               the path; or ``None`` if no path from start to goal exists.
    """
    if goal_test(start_state):
        return [start_state]
    agenda = PriorityQueue()
    agenda.push(SearchNode(start_state, None, cost=0), heuristic(start_state))
    expanded = set()
    while len(agenda) > 0:
        priority, parent = agenda.pop()
        if parent.state not in expanded:
            expanded.add(parent.state)
            if goal_test(parent.state):
                return parent.path()
            for child_state, cost in successors(parent.state):
                child = SearchNode(child_state, parent, parent.cost+cost)
                if child_state not in expanded:
                    agenda.push(child, child.cost+heuristic(child_state))
    return None
