"""
Class and some supporting functions for representing and manipulating system functionals. 
"""

import imp
import math

from . import poly
imp.reload(poly)

from . import sim
imp.reload(sim)

class SystemFunctional:
    """
    Represent a system functional as a ratio of polynomials in :math:`{\\cal R}`
    """
#!
    def __init__(self, numeratorPoly, denominatorPoly):
        """
        :param numeratorPoly: :py:class:`lib601.poly.Polynomial` reresenting
                              the system functional's numerator, as a
                              polynomial in :math:`{\\cal R}`
        :param denominatorPoly: :py:class:`lib601.poly.Polynomial` reresenting
                                the system functional's denominator, as a
                                polynomial in :math:`{\\cal R}`
        """
        self.numerator = numeratorPoly
        """L{Polynomial<poly.Polynomial>} in R representing the numerator"""
        self.denominator = denominatorPoly
        """L{Polynomial<poly.Polynomial>} in R representing the denominator"""

    def poles(self):
        """
        :returns: a list of the poles of the system
        """
        # Find the poles of a system functional by substituting
        # 1/z for R in both numerator and denominator, and then
        # finding the roots of the denominator polynomial of z.

        # We can find the coefficients of the z polynomial by reversing the
        # order of the coefficients of the R polynomial.
        # However, we must make the numerator and denominator coefficient lists
        # have the same length (by padding with zeros) to avoid introducing
        # false roots (at zero).

        # create equal-length lists of coefficients corresponding to polynomials in z
        num = []
        den = []
        for i in range(1+max(self.numerator.order,self.denominator.order)):
            num.append(self.numerator.coeff(i))
            den.append(self.denominator.coeff(i))
        # cancel poles at zero with zeros at zero
        while len(num) > 0 and num[-1]==den[-1]==0:
            num.pop()
            den.pop()
        while len(den)>0 and den[0]==0:
            den.pop(0)
        return poly.Polynomial(den).roots()

    def pole_magnitudes(self):
        """
        :returns: a list of the magnitudes of the poles of the system
        """
        return [abs(r) for r in self.poles()]

    def dominant_pole(self):
        """
        :returns: the pole with the largest magnitude
        """
        return max(self.poles(), key=abs)

    def simulator(self):
        """
        :returns: An instance of :py:class:`lib601.sim.DifferenceEquationSim`
                  which is equivalent to this system (started from rest)
        """
        num = [self.numerator.coeff(i) for i in range(self.numerator.order+1)]
        denom = [self.denominator.coeff(i) for i in range(self.denominator.order+1)]

        assert denom[0] != 0, "Coefficient of y[n] must be non-zero"

        norm = float(denom[0])

        dCoeffs = [i/norm for i in num]
        cCoeffs = [-i/norm for i in denom[1:]]
        return sim.DifferenceEquationSim(dCoeffs,cCoeffs)

    def __str__(self):
        return 'SystemFunctional(%r,%r)' % (self.numerator,self.denominator)

    __repr__ = __str__

#!
######################################################################
# A symmetric syntax for combining SystemFunctionals
######################################################################
#!
def Cascade(sf1, sf2):
#!    pass
    """
    :param sf1: :py:class:`SystemFunctional`
    :param sf2: :py:class:`SystemFunctional`
    :returns: :py:class:`SystemFunctional` representing the cascade of ``sf1`` and ``sf2``
    """
    return SystemFunctional(sf1.numerator * sf2.numerator,
                          sf1.denominator * sf2.denominator)
#!
#!
def FeedforwardAdd(sf1, sf2):
    """
    :param sf1: :py:class:`SystemFunctional`
    :param sf2: :py:class:`SystemFunctional`
    :returns: :py:class:`SystemFunctional` representing the sum of ``sf1`` and
              ``sf2``; this models the situation when the two component systems
              have the same input and the output of the whole system is the sum
              of the outputs of the components.
    """
    return SystemFunctional(sf1.numerator*sf2.denominator + \
                          sf2.numerator*sf1.denominator,
                          sf1.denominator * sf2.denominator)
def FeedforwardSubtract(sf1, sf2):
    """
    :param sf1: :py:class:`SystemFunctional`
    :param sf2: :py:class:`SystemFunctional`
    :returns: :py:class:`SystemFunctional` representing the difference of
              ``sf1`` and ``sf2``; this models the situation when the two
              component systems have the same input and the output of the whole
              system is the output of the first component minus the output of
              the second component.
    """
    return SystemFunctional(sf1.numerator*sf2.denominator - \
                          sf2.numerator*sf1.denominator,
                          sf1.denominator * sf2.denominator)
   

def FeedbackSubtract(sf1, sf2=None):
    """
    :param sf1: :py:class:`SystemFunctional`
    :param sf2: :py:class:`SystemFunctional`
    :returns: :py:class:`SystemFunctional` representing the result of feeding
              the output of ``sf1`` back, with (optionally) ``sf2`` on the
              feedback path, subtracting it from the input, and feeding the
              resulting signal into C{sf1}.  This situation can be
              characterized with Black's formula. 
    """
    if sf2 == None:
        sf2 = SystemFunctional(poly.Polynomial([1]),
                             poly.Polynomial([1]))
    (n1, d1) = (sf1.numerator, sf1.denominator)
    (n2, d2) = (sf2.numerator, sf2.denominator)
    return SystemFunctional(n1 * d2, d1 * d2 + n1 * n2)


def FeedbackAdd(sf1, sf2=None):
    """
    :param sf1: :py:class:`SystemFunctional`
    :param sf2: :py:class:`SystemFunctional`
    :returns: :py:class:`SystemFunctional` representing the result of feeding
              the output of ``sf1`` back, with (optionally) ``sf2`` on the
              feedback path, adding it to the input, and feeding the resulting
              signal into C{sf1}.
    """
    if sf2 == None:
        sf2 = SystemFunctional(poly.Polynomial([1]),
                             poly.Polynomial([1]))
    (n1, d1) = (sf1.numerator, sf1.denominator)
    (n2, d2) = (sf2.numerator, sf2.denominator)
    return SystemFunctional(n1 * d2, d1 * d2 - n1 * n2)

def Sum(sf1, sf2):
    """
    :param sf1: :py:class:`SystemFunctional`
    :param sf2: :py:class:`SystemFunctional`
    :returns: :py:class:`SystemFunctional` representing the system that sums
              the outputs of the two systems
    """
    (n1, d1) = (sf1.numerator, sf1.denominator)
    (n2, d2) = (sf2.numerator, sf2.denominator)
    return SystemFunctional(n1 * d2 + n2 * d1, d1 * d2)

######################################################################
##    Primitive SF's
######################################################################

def Gain(k):
    """
    :param k: gain parameter
    :returns: :py:class:`SystemFunctional` representing a system that
              multiplies the input signal by ``k``.
    """
    return SystemFunctional(poly.Polynomial([k]),poly.Polynomial([1]))

def R():
    """
    :returns: :py:class:`SystemFunctional` representing a system that delays
              the input signal by one step.
    """
    return SystemFunctional(poly.Polynomial([1, 0]),poly.Polynomial([1]))
