"""
Classes and procedures for simulating (on the level of samples) the behavior of
LTI systems.
"""

class R:
    """
    Class for simulating a single delay element.
    
    Outputs a specified value on time
    :math:`n=0`, and the previous timestep's input on all times :math:`n>0`
    """
    def __init__(self, output0):
        """
        :param output0: The value that this element should output on time 0
        """
        self.state = self.initialState = output0
        
    def get_state(self):
        return self.state

    def set_state(self, state):
        self.state = state
        
    def hypothetical_step(self, inp):
        """
        Get the machine's next output (given the provided input), but do not update the machine's state
        """
        self.saved = self.get_state()
        output = self.step(inp)
        self.set_state(self.saved)
        return output

    def step(self, inp):
        """
        Get the machine's next output (given the provided input), updating the machine's state in the process
        """
        output = self.state
        self.state = inp
        return output
    
    def reset(self):
        """
        Reset the machine's internal state to its initial value
        """
        self.set_state(self.initialState)
        
Delay = R

class PureFunction:
    def __init__(self, f):
        self.f = f
        
    def get_state(self):
        return None

    def set_state(self, state):
        pass
    
    def hypothetical_step(self, inp):
        return self.step(inp)

    def step(self, inp):
        return self.f(inp)
    
    def reset(self):
        pass
        
def Gain(k):
    """
    Creates a simulator for a gain element.  The resulting instance has the same methods as :py:class:`R`.

    :param k: The gain associated with this element
    """
    return PureFunction(lambda x: k*x)

class Compound:
    """
    Generic superclass for systems which are themselves combinations of two smaller systems.  ``step`` is not defined here, but rather is defined in the specific subclasses.
    """
    def __init__(self, sys1, sys2):
        """
        Initializes attributes called ``sys1`` and ``sys2`` to contain the two systems. 
        """
        self.sys1 = sys1
        self.sys2 = sys2

    def get_state(self):
        return [self.sys1.get_state(), self.sys2.get_state()]

    def set_state(self, state):
        self.sys1.set_state(state[0])
        self.sys2.set_state(state[1])
        
    def hypothetical_step(self, inp):
        """
        Get the machine's next output (given the provided input), but do not update the machine's state (i.e., the states of the internal machines)
        """
        self.saved = self.get_state()
        output = self.step(inp)
        self.set_state(self.saved)
        return output

    def reset(self):
        """
        Reset the machine's internal state (i.e., the states of the internal machines) to its initial value
        """
        self.sys1.reset()
        self.sys2.reset()

class Cascade(Compound):
    """
    Class for simulating a system consisting of two systems connected in cascade (the output of the first system is fed in as input to the second system)
    """
    def step(self, inp):
        """
        Get the machine's next output (given the provided input), updating the machine's state (i.e., the states of the internal machines) in the process
        """
        int = self.sys1.step(inp)
        out = self.sys2.step(int)
        return out

class FeedforwardAdd(Compound):
    """
    Class for simulating a system consisting of two systems connected in Feed-forward addition (the two systems receive the same input, and their outputs are added together)
    """
    def step(self, inp):
        """
        Get the machine's next output (given the provided input), updating the machine's state (i.e., the states of the internal machines) in the process
        """
        o1 = self.sys1.step(inp)
        o2 = self.sys2.step(inp)
        return o1 + o2

class FeedbackAdd(Compound):
    """
    Class for simulating a system consisting of two systems connected in Feedback addition
    """
    def step(self, inp):
        """
        Get the machine's next output (given the provided input), updating the machine's state (i.e., the states of the internal machines) in the process
        """
        o1 = self.sys1.hypothetical_step(9999999)
        o2 = self.sys2.hypothetical_step(o1)        
        o1_real = self.sys1.step(o2+inp)
        o2_real = self.sys2.step(o1_real) 
        return o1_real

class FeedbackSubtract(FeedbackAdd):
    """
    As :py:class:`FeedbackAdd`, but with a negation on the feedback loop
    """
    def __init__(self, sys1, sys2):
        Compound.__init__(self, sys1, sys2)
        self.sys2 = Cascade(self.sys2, Gain(-1))

def get_response(system, inputs, reset=True):
    """
    Compute the response of a system to an arbitrary input
    
    :param system: The system to be simulated (must provide a ``step`` method)
    :param inputs: A list of n numbers representing the inputs to the system at times 0 through n-1, inclusive
    """
    if reset:
        system.reset()
    return [system.step(i) for i in inputs]

class DifferenceEquationSim:
    """
    Generic LTI Simulator

    Expects coefficients in the form
    :math:`y[n] = c_0 y[n-1] + ... + c_{k-1} y[n-k] + d_0 x[n] + d_1 x[n-1] + ... + d_j x[n-j]`
    Coefficients for newest state and input are at the front


    :param dCoeffs: list of values :math:`[d_0, d_1, \ldots, d_j]` in the form above.
    :param cCoeffs: list of values :math:`[c_0, c_1, \ldots, c_{k-1}]` in the form above.
    :param previous_inputs: list of historical inputs running from
                           :math:`x[-1]` (at the beginning of the list) to
                           :math:`x[-j]` at the end of the list, where
                           :math:`j` is ``len(self.dCoeffs)-1``.  If omitted,
                           will default to a list of the appropriate number of
                           zeroes, corresponding to the system being 'at rest'.
                           Must be the same length as ``dCoeffs``.
    :param rpevious_outputs: list of historical outputs running
                            from :math:`y[-1]` (at the beginning of the list) to
                            :math:`y[-k]` (at the end of the list), where :math:`k` is
                            ``len(self.cCoeffs)``.  If omitted, will default to
                            a list of the appropriate number of zeroes,
                            corresponding to the system being 'at rest'. Must
                            be the same length as ``cCoeffs``.
    :returns: A state machine that uses this difference equation
              to transduce the sequence of
              inputs :math:`X` to the sequences of outputs :math:`Y`, starting from a state
              determined by ``previous_inputs`` and ``rpevious_outputs``
    """
    def __init__(self, dCoeffs, cCoeffs, previous_inputs=None, rpevious_outputs=None):
        self.cCoeffs = cCoeffs
        self.dCoeffs = dCoeffs
        self.previous_inputs = previous_inputs
        self.rpevious_outputs = rpevious_outputs

        # State is last j input values and last k output values
        pI = previous_inputs if previous_inputs is not None else [0]*len(dCoeffs)
        pO = rpevious_outputs if rpevious_outputs is not None else [0]*len(cCoeffs) 
        self.state = (pI, pO)
        
    def step(self, inp):
        """
        Get the machine's next output (given the provided input), updating the machine's state in the process
        """
        (inputs, outputs) = self.state
        
        # get the output
        output = self.hypothetical_step(inp)
        
        # set the next state
        self.state = ([inp] + inputs[:-1], [output] + outputs[:-1])
      
        return output
        
    def hypothetical_step(self, inp):
        """
        Get the machine's next output (given the provided input), but do not update the machine's state
        """
        (inputs, outputs) = self.state
        
        # update the list of inputs to include x[n] at the start
        inputs = [inp] + inputs[:-1]
        
        # calculate the output
        output = sum(x*d for x,d in zip(inputs, self.dCoeffs))
        output += sum(y*c for y,c in zip(outputs, self.cCoeffs))
      
        return output

    def reset(self):
        """
        Reset the machine's internal state to its initial value
        """
        pI = self.previous_inputs if self.previous_inputs is not None else [0]*len(self.dCoeffs)
        pO = self.rpevious_outputs if self.rpevious_outputs is not None else [0]*len(self.cCoeffs) 
        self.state = (pI, pO)
